/* CVS: CobaltVault SERVER:\\halo TREE:\abode_mainline
 *       _    ____   ___  ____  _____ 
 *      / \  | __ ) / _ \|  _ \| ____|      Advanced
 *     / _ \ |  _ \| | | | | | |  _|        Behavior
 *    / ___ \| |_) | |_| | |_| | |___       Oriented
 *   /_/   \_\____/ \___/|____/|_____|      Design
 *         www.cobaltsoftware.net           Environment
 *
 * PRODUCED FOR:      University of Bath / Boeing
 * PAYMENT:           On Delivery   
 * LICENSING MODEL:   Unrestricted distribution (Post Delivery)
 * COPYRIGHT:         Client retains copyright.
 *
 * This program and all the software components herein are
 * released as-is, without warranties regarding function,
 * correctness or any other aspect of the components.
 * Steven Gray, Cobalt Software, it's subcontractors and
 * successors may not be held liable for any damage caused 
 * to computers, business or other property through use of 
 * or misuse of this software.
 *
 * Upon redistribution of the program, all notices of
 * copyrights, both of the software provider and the 
 * client must be retained.
 */ 
package abode.lib;
import java.io.*;
import java.util.*;

/**
 * The DotLapReader reads in lisp-formatted .lap files and then produces a completed 
 * LearnableActionPattern object representing the contents of the file.
 *
 * @author  CobaltSoftware (abode.devteam@cobaltsoftware.net)
 * @version 1.0
 */
public class DotLapReader implements ILAPReader {
    /**
     * Is the specified file readable to this class?
     *
     * @param fileName Path to the file to read
     * @return True if readable, false if not
     **/
    public boolean canRead(String strFileName) {
        // If we can load this file without an exception from the
        // DotLapReader then we're capable of loading this file. 
        // This could perhaps be done more efficiently.
        try {
            load(strFileName);
            return true;
        } catch (Exception e) {
            return false;
        }
    }

    /**
     * Read the contents of a text file
     * 
     * @param strFile File to load 
     * @return Contents of file
     **/
    public String getFileContents(String strFileName) {
        try {
            BufferedReader brInputFile = new BufferedReader(new FileReader(strFileName));
            String strFileContent = "", temp = "";
            while ((temp = brInputFile.readLine()) != null)
                strFileContent += temp + "\n";
            brInputFile.close();
            return strFileContent;
        }
        catch (Exception e) {
            return "Unable to load file.";
        }
    }
    
    /**
     * Load the specified file and get an action pattern
     * 
     * @param fileName File to load
     * @return Loaded file in the form of an object model
     **/
    public LearnableActionPattern load(String strFileName) throws FileNotFoundException, IOException, Exception {
        // Open our file for reading
        BufferedReader brInputFile = new BufferedReader(new FileReader(strFileName));
        
        try {
            // Load each line of the file into a string
            String strFileContent = "", temp = "";
            while ((temp = brInputFile.readLine()) != null)
                strFileContent += temp + "\n";
            
            // Parse the Lisp document object model
            LispBlob blob = new LispBlob(strFileContent);
            
            ArrayList elements = new ArrayList();
            if (!blob.isList())
                throw new Exception("Invalid file structure (Initial element not a list.). Cannot parse this file.");
            
            // Default to blank documentation
            Documentation documentation = new Documentation();
            
            // Iterate over the list and see if we can find the documentation
            Iterator children = blob.getIterator();
    
            // Carry on scanning
            while (children.hasNext()) {
                // Get this child of the list
                LispBlob child = (LispBlob) children.next();

                // Add it to our children list after parsing it
                Object element = parseElement(child);
                if (element instanceof Documentation)
                    documentation = (Documentation) element;
                else
                    elements.add(element);
            }
            
            return new LearnableActionPattern(elements, documentation);
        } 
        finally {
            brInputFile.close();
            brInputFile = null;
        }
    }
    
    /**
     * Parse a construct from the file
     **/
    private Object parseElement(LispBlob block) throws Exception {
        // All of the various constructs must be lists, so we exception
        // if there is no list inside this list.
        if (!block.isList())
            throw new Exception("Invalid file structure (Elements must be composed of lists!). Cannot parse this file.");
        
        // Copy out each of the children into an arraylist for processing as an array
        ArrayList children = block.toList();
        
        // The first item of the first child will be a constant
        LispBlob first = (LispBlob) children.get(0);
        
        // This constant is our type identifier
        String strType = first.getText();
        
        if (strType.equals("documentation"))
            return parseDocumentation(children);
        if (strType.equals("AP"))
            return parseActionPattern(children);
        if (strType.equals("C"))
            return parseCompetence(children);
        if (strType.equals("DC"))
            return parseDriveCollection(children, false);
        if (strType.equals("RDC"))
            return parseDriveCollection(children, true);
        return null;
    }
    
    /**
     * Parse documentation
     *
     * @param elements Arraylist containg documentation elements
     * @return Documentation construct
     **/
    private Documentation parseDocumentation(ArrayList elements) {
        return new Documentation( ((LispBlob) elements.get(1)).getText().replaceAll("\"", ""), ((LispBlob) elements.get(2)).getText().replaceAll("\"", ""), ((LispBlob) elements.get(3)).getText().replaceAll("\"", "") );
    }
    
    /**
     * Parse a DC or (R) DC block
     **/
    private DriveCollection parseDriveCollection(ArrayList elements, boolean realTime) throws Exception {
        // Get the name, parse the goal and the get the list of drive lists
        String strName = ((LispBlob) elements.get(1)).getText();
        ArrayList goal = parseGoal((LispBlob) elements.get(2));
        ArrayList lists = ((LispBlob) elements.get(3)).toList();
        
        // Remove the constant "Drives" which is at the start of the list
        lists.remove(0);
        
        // For each list of drive elements
        ArrayList elementLists = new ArrayList();
        Iterator listIterator = lists.iterator();
        while (listIterator.hasNext()) {
            ArrayList subList = new ArrayList();
            Iterator subListIterator = ((LispBlob) listIterator.next()).toList().iterator();
            while (subListIterator.hasNext())
                subList.add(parseDriveElement((LispBlob) subListIterator.next()));
            elementLists.add(subList);
        }
        
        JAbode.writeEnvironmentLine("Parsed drive collection (" + strName + ")");
        return new DriveCollection(strName, realTime, goal, elementLists);
    }
    
    /**
     * Parse a single drive element
     **/
    private DriveElement parseDriveElement(LispBlob blob) throws Exception {
        // Convert to list, extract name, trigger, action name and whether or not there's a timeout
        ArrayList list = blob.toList();
        
        String strName = ((LispBlob) list.get(0)).getText();
        ArrayList trigger = parseTrigger((LispBlob) list.get(1));
        String strAction = ((LispBlob) list.get(2)).getText();
        
       
        JAbode.writeEnvironmentLine("Parsed Drive Element (" + strName + ")");
        
        // DriveElement(String name, ArrayList trigger, String action) 
        if (list.size() == 4) {
            TimeUnit time = parseTimeUnit((LispBlob) list.get(3));
            return new DriveElement(strName, trigger, strAction, time);
        }
        
        return new DriveElement(strName, trigger, strAction);
    }
    
    /**
     * Parse a competence block
     **/
    private Competence parseCompetence(ArrayList elements) throws Exception {
        // Extract some infomration from the list
        String      strName      = ((LispBlob) elements.get(1)).getText();
        TimeUnit    timeUnit     = parseTimeUnit((LispBlob) elements.get(2));
        ArrayList   goal         = parseGoal((LispBlob) elements.get(3));
        LispBlob elementLists = (LispBlob) elements.get(4);

        // Recurse the lists of lists of competence elements and build our structure
        ArrayList compElements = new ArrayList();
        Iterator it = elementLists.getIterator();
        
        // Skip over constant value "elements"
        if (it.hasNext())
            it.next();
        
        // For each list of lists
        while (it.hasNext()) {
            // Iterate over the list inside
            LispBlob list = (LispBlob) it.next();
            Iterator subList = list.toList().iterator();

            ArrayList buildList = new ArrayList();
            while (subList.hasNext()) {
                // Produce the competence
                CompetenceElement ce = parseCompetenceElement((LispBlob) subList.next());
                buildList.add(ce);
            }
            compElements.add(buildList);
        }
        
        JAbode.writeEnvironmentLine("Parsed competence (" + strName + ")");
        return new Competence(strName, timeUnit, goal, compElements);    
    }
    
    /**
     * Parse a competence element
     **/
    private CompetenceElement parseCompetenceElement(LispBlob blob) throws Exception {
        ArrayList list = blob.toList();
        
        // Get the name, trigger list and resulting action
        String strName = ((LispBlob) list.get(0)).getText();
        ArrayList alTrigger = parseTrigger((LispBlob) list.get(1));
        String action = ((LispBlob) list.get(2)).getText();
       
        JAbode.writeEnvironmentLine("Parsed competence element (" + strName + ")");
        
        if (list.size() == 3)
            return new CompetenceElement(strName, alTrigger, action);
        
        if (list.size() == 4) {
            int iRet = Integer.parseInt(((LispBlob) list.get(3)).getText());
            return new CompetenceElement(strName, alTrigger, action, iRet);
        }
        
        throw new Exception("Invalid number of elements for competence element!");
    }
    
    /**
     * Parse a trigger block
     **/
    private ArrayList parseTrigger(LispBlob blob) throws Exception {
        ArrayList source = blob.toList();
        LispBlob blob2 = (LispBlob) source.get(1);
        if (!blob2.isList())
            return new ArrayList(); // "nil"
        return parseActionElementList(blob2);
    }
    
    /**
     * Parse a goal block - This is essentially an actionelement list with a bit of wrapping
     **/
    private ArrayList parseGoal(LispBlob blob) throws Exception {
        ArrayList list = blob.toList();
        return parseActionElementList((LispBlob) list.get(1));
    }
    
    
    /**
     * Parse an arraylist of elements that represents an action pattern
     **/
    private ActionPattern parseActionPattern(ArrayList elements) throws Exception {
        // Our name is the second element of the list
        String strName = ((LispBlob) elements.get(1)).getText();
 
        // Our timeout can be parsed from the third element of the list
        TimeUnit timeOut = parseTimeUnit((LispBlob) elements.get(2));
        
        // Our list of ActionElements is the fourth element of the list
        LispBlob ael = (LispBlob) elements.get(3);
        
        // Parse each element of this list
        ArrayList children = parseActionElementList(ael);
       
        JAbode.writeEnvironmentLine("Parsed action pattern named:" + strName);
        return new ActionPattern(strName, timeOut, children);
    }
    
    /**
     * Parse a list of action elements
     **/
    private ArrayList parseActionElementList(LispBlob blob) throws Exception {
        ArrayList children = new ArrayList();
        Iterator it = blob.getIterator();
        while (it.hasNext())
            children.add(parseActionElement((LispBlob) it.next()));
       return children;
    }
    
    /**
     * Parse an action element from a blob of lisp
     **/
    private ActionElement parseActionElement(LispBlob blob) throws Exception {
        // Simple case, we're just a name
        if (!blob.isList())
            return new ActionElement(false, blob.getText());
        
        ArrayList list = blob.toList();
        String name = ((LispBlob) list.get(0)).getText();
        switch (list.size()) {
            case 3:
                String pred = ((LispBlob) list.get(2)).getText();
                String val =((LispBlob) list.get(1)).getText();
                return new ActionElement(name, val, pred);
            case 2:
                String val2 = ((LispBlob) list.get(1)).getText();
                return new ActionElement(name, val2);
            default:
                return new ActionElement(true, name);
        }
    }
    
    /**
     * Parse a unit of time from a lisp blob
     **/
    private TimeUnit parseTimeUnit(LispBlob blob) throws Exception {
        ArrayList list = blob.toList();
        String unit = ((LispBlob) list.get(0)).getText();
        String val = ((LispBlob) list.get(1)).getText();
        return new TimeUnit(unit, Double.parseDouble(val));
    }
}
